"""
Bunyan Formatter.

Provides logging compatibility with Bunyan standard and enceforth CLI.
"""
import datetime
import json
import logging
import re
import socket
import traceback

from inspect import istraceback

#Support order in python 2.7 and 3
try:
  from collections import OrderedDict
except ImportError:
  pass

def object_startswith(key, value):
  return hasattr(key, 'startswith') and key.startswith(value)

def merge_record_extra(record, target, reserved = []):
  """
  Merges extra attributes from LogRecord object into target dictionary.

  :param record: logging.LogRecord
  :param target: dict to update
  :param reserved: dict or list with reserved keys to skip
  """
  new_values = {
    key: value
    for key, value in record.__dict__.items()
      if (key not in reserved and not object_startswith(key, '_'))
  }

  target.update(new_values)

  return target

def get_json_handler(datefmt):
  def handler(obj):
    '''Prints dates in ISO format'''
    if isinstance(obj, datetime.datetime):
      if obj.year < 1900:
        # strftime do not work with date < 1900
        return obj.isoformat()

      return obj.strftime(datefmt or '%Y-%m-%dT%H:%M')
    elif isinstance(obj, datetime.date):
      return obj.isoformat()
    elif isinstance(obj, datetime.time):
      return obj.strftime('%H:%M')
    elif istraceback(obj):
      tb = ''.join(traceback.format_tb(obj))

      return tb.strip()
    elif isinstance(obj, Exception):
      return "Exception: %s" % str(obj)

    return str(obj)

  return handler


class BunyanFormatter(logging.Formatter):
  """
  Bunyan Formatter.

  Implements a logging Formatter by extending jsonlogger.JsonFormatter
  to use bunyan's standard names and values.
  """
  def __init__(self, *args, **kwargs):
    """
    Defined default log format.
    """
    self._required_fields = [
      'asctime',
      'exc_info',
      'levelno',
      'message',
      'name',
      'process',
    ]
    self._skip_fields = self._required_fields[:]
    self._skip_fields += [
      'args',
      'created',
      'exc_text',
      'filename',
      'funcName',
      'levelname',
      'lineno',
      'module',
      'msecs',
      'pathname'
      'processName',
      'relativeCreated',
      'stack_info',
      'thread',
      'threadName',
    ]

    log_format = lambda x: ["%({0:s})".format(i) for i in x]
    logging.Formatter.__init__(self, ' '.join(log_format(self._required_fields)), "%Y-%m-%dT%H:%M:%SZ", *args, **kwargs)
    self.json_default = get_json_handler(self.datefmt)

  def add_fields(self, log_record, record, message_dict):
    """
    Override this method to implement custom logic for adding fields.
    """
    for field in self._required_fields:
      log_record[field] = record.__dict__.get(field)

    log_record.update(message_dict)
    merge_record_extra(record, log_record, reserved = self._skip_fields)

  def process_log_record(self, log_record):
    """
    Override this method to implement custom logic
    on the possibly ordered dictionary.
    """
    return log_record

  def jsonify_log_record(self, log_record):
    """
    Returns a json string of the log record.
    """
    return json.dumps(log_record, default = self.json_default)

  def format(self, record):
    """
    Formats a log record and serializes to json
    """
    message_dict = {}

    if isinstance(record.msg, dict):
      message_dict = record.msg
      record.message = None
    else:
      record.message = record.getMessage()
    # only format time if needed
    if "asctime" in self._required_fields:
      record.time = self.formatTime(record, self.datefmt)

    # Display formatted exception, but allow overriding it in the
    # user-supplied dict.
    if record.exc_info and not message_dict.get('exc_info'):
      message_dict['exc_info'] = self.formatException(record.exc_info)

    try:
      log_record = OrderedDict()
    except NameError:
      log_record = {}

    self.add_fields(log_record, record, message_dict)
    log_record = self.process_log_record(log_record)

    return self.jsonify_log_record(log_record)

  def process_log_record(self, log_record):
    """
    Bunyanize log_record:
      - Renames python's standard names by bunyan's.
      - Add hostname and version (v).
      - Normalize level (+10).
    """
    #Add hostname
    log_record['hostname'] = socket.gethostname()
    log_record['level'] = log_record['levelno'] + 10

    if 'message' in log_record and log_record['message']:
      log_record['msg'] = log_record['message']
    else:
      log_record['msg'] = ""

    log_record['pid'] = log_record['process']
    log_record['v'] = 0

    if not log_record['exc_info']:
      del log_record['exc_info']

    del log_record['asctime']
    del log_record['levelno']
    del log_record['message']
    del log_record['process']

    return log_record



import datetime
import json
import logging
import sys
import traceback
import unittest

try:
  from StringIO import StringIO
except ImportError:
  # Python 3 Support
  from io import StringIO

sys.path.append('../bunyan')
import bunyan


class TestJsonLogger(unittest.TestCase):
  def setUp(self):
    self.logBuffer = StringIO()

    self.logger = logging.getLogger('bunyan-test')
    self.logger.setLevel(logging.DEBUG)

    self.logHandler = logging.StreamHandler(self.logBuffer)
    self.logger.addHandler(self.logHandler)

  def testDefaultFormat(self):
    fr = bunyan.BunyanFormatter()
    self.logHandler.setFormatter(fr)

    msg = "testing logging format"
    self.logger.info(msg)
    logJson = json.loads(self.logBuffer.getvalue())

    self.assertEqual(logJson["msg"], msg)

  def testFormatKeys(self):
    supported_keys = [
      'hostname',
      'level',
      'msg',
      'name',
      'pid',
      'time',
      'v'
    ]
    fr = bunyan.BunyanFormatter()
    self.logHandler.setFormatter(fr)

    msg = "testing bunyan"
    self.logger.info(msg)
    log_msg = self.logBuffer.getvalue()
    log_json = json.loads(log_msg)

    for supported_key in supported_keys:
      self.assertIn(supported_key, log_json)

  def testLogADict(self):
    fr = bunyan.BunyanFormatter()

    self.logHandler.setFormatter(fr)

    msg = {
      "the": "dict",
      "has": 1,
      5: 1,
      "and": {"nested": "properties"}
    }

    self.logger.info(msg)
    logJson = json.loads(self.logBuffer.getvalue())

    self.assertEqual(logJson.get("the"), msg["the"])
    self.assertEqual(logJson.get("has"), msg["has"])
    self.assertEqual(logJson.get("5"), msg[5])
    self.assertEqual(logJson.get("and"), msg["and"])
    self.assertEqual(logJson["msg"], "")

  def testLogExtra(self):
    fr = bunyan.BunyanFormatter()

    self.logHandler.setFormatter(fr)

    extra = {
      "the": "dict",
      "has": 1,
      5: 1,
      "and": {
        "nested": "properties"
      }
    }

    self.logger.info("yo", extra = extra)
    logJson = json.loads(self.logBuffer.getvalue())

    self.assertEqual(logJson.get("the"), extra["the"])
    self.assertEqual(logJson.get("has"), extra["has"])
    self.assertEqual(logJson.get("5"), extra[5])
    self.assertEqual(logJson.get("and"), extra["and"])
    self.assertEqual(logJson["msg"], "yo")

  def testJsonDefaultEncoder(self):
    fr = bunyan.BunyanFormatter()

    self.logHandler.setFormatter(fr)

    msg = {
      "dateone": datetime.datetime(1999, 12, 31, 23, 59),
      "datetwo": datetime.datetime(1900, 1, 1)
    }

    self.logger.info(msg)
    logJson = json.loads(self.logBuffer.getvalue())

    self.assertEqual(logJson.get("dateone"), "1999-12-31T23:59:00Z")
    self.assertEqual(logJson.get("datetwo"), "1900-01-01T00:00:00Z")

  def testJsonCustomLogicAddsField(self):

    class LeBunyanFormatter(bunyan.BunyanFormatter):
      def process_log_record(self, log_record):
        log_record["addit"] = "added"
        return super(LeBunyanFormatter, self).process_log_record(log_record)

    self.logHandler.setFormatter(LeBunyanFormatter())
    self.logger.info("yo")

    logJson = json.loads(self.logBuffer.getvalue())
    self.assertEqual(logJson.get("addit"), "added")

  def testExcInfo(self):
    fr = bunyan.BunyanFormatter()

    self.logHandler.setFormatter(fr)
    try:
      raise Exception('letest')
    except Exception:
      self.logger.exception("yo")

      expected_value = traceback.format_exc()
      # Formatter removes trailing new line
      if expected_value.endswith('\n'):
        expected_value = expected_value[:-1]

    logJson = json.loads(self.logBuffer.getvalue())
    self.assertEqual(logJson.get("exc_info"), expected_value)



from setuptools import find_packages
from distutils.core import setup

from io import open
try:
  import pypandoc
  long_description = pypandoc.convert('README.md', 'rst')
  long_description += pypandoc.convert('CHANGELOG.md', 'rst')
except(IOError, ImportError, OSError):
  long_description = open('README.md').read()
  long_description += open('CHANGELOG.md').read()

setup(
  name = 'bunyan',
  version = '0.1.1',
  author = 'Jorge Alpedrinha Ramos',
  author_email = 'python@uphold.com',
  packages = find_packages(),
  package_data = { '': ['*.yml']},
  license = 'LICENSE',
  description = 'Bunyan python logger',
  url = 'https://www.github.com/uphold/python-bunyan/',
  long_description = long_description
)
